/*
 * File         : Choleski.java
 * Author       : Guansong Zhang
 * Created      : Sat Jan 24 17:29:26 1998
 * Revision     : $Revision: 1.5 $
 * Updated      : $Date: 1998/02/20 17:01:44 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

import spmd.*;

class Choleski {

  static final int n=@size@;

  public static void main(String argv[]) {
    SpmdEnv spmd=new SpmdEnv(argv);

    Procs1 p = new Procs1(@pnum@);

    if (p.on()) {

      Range x = new Range(n);
      Range y = new CyclicRange(n, p.dim(0));

      Array2Float a = new Array2Float(x,y, spmd.apg);

      Index ic = new Index(x);
      Index jc = new Index(y);

      ic.dim().over();
      for (ic.begin(); ic.test(); ic.next()) {
	jc.dim().over();
	for (jc.begin(); jc.test(); jc.next()) {
	  if (jc.idx()>=ic.idx())
	    a.data[a.element(ic,jc)]=ic.idx()+1;
	  else
	    a.data[a.element(ic,jc)]=jc.idx()+1;
	} jc.dim().revo();
      } ic.dim().revo();

      //      Adlib.rprintln("A = ");
      //      Adlib.aprintf("%7.3f%"+n+"N", a);

      Array1Float b = new Array1Float(x, spmd.apg);

      /*
      Location i;
      Location j;

      i = x.location(0);
      j = y.location(0);

      if (i.at()) {
	if (j.at()) {
	  a.data[a.element(i,j)]= (float) Math.sqrt(a.data[a.element(i,j)]);
	} j.ta();
      } i.ta();
      
      for (int k=0; k<n-1; k++) {

	ic = new Index(x.subrng(n-k-1, k+1));
	ic.dim().over();
	for (ic.begin(); ic.test(); ic.next()) {
	  j = y.location(k);
	  if (j.at()) {
	    a.data[a.element(ic,j)] /= a.data[a.element(x.location(k),j)];
	  } j.ta();
	} ic.dim().revo();

	Adlib.remap(b.section(x.subrng(n-k-1,k+1)), 
		    a.section(x.subrng(n-k-1,k+1), y.location(k)));

	jc = new Index (y.subrng(n-k-1, k+1));
	jc.dim().over();
	for (jc.begin(); jc.test(); jc.next()) {

	  ic = new Index(x.subrng(n-k-1-jc.idx(), k+1+jc.idx()));
	  ic.dim().over();
	  for (ic.begin(); ic.test(); ic.next()) {

	    a.data[a.element(ic,jc)] -= b.data[b.element(ic)]*
	      b.data[b.element(x.location(k+1+jc.idx()))];

	  } ic.dim().revo();

	} jc.dim().revo();

	i = x.location(k+1);
	j = y.location(k+1);

	if (i.at()) {
	  if (j.at()) {
	    a.data[a.element(i,j)]= (float) Math.sqrt(a.data[a.element(i,j)]);
	  } j.ta();
	} i.ta();

	Adlib.rprintln(k+"th A = ");
	Adlib.aprintf("%7.3f%"+n+"N", a);

      }

      */

      Clock.start();

      Location i;
      Location j;

      float [] a_ptr=a.data;
      float [] b_ptr=b.data;

      final int a0_bas=a.stride[0].disp(x.bas());
      final int a0_stp=a.stride[0].disp(x.str());

      final int b0_bas=b.stride[0].disp(x.bas());
      final int b0_stp=b.stride[0].disp(x.str());

      for(int k=0; k<n-1; k++) {
	j=y.location(k);

	if (j.at()) {
	  final int a1_off=a.stride[1].offset(j);
	  final float d = (float)Math.sqrt(a_ptr[a0_bas+a0_stp*k+a1_off]);
	  a_ptr[a0_bas+a0_stp*k+a1_off]=d;

	  for (int s=k+1; s<n; s++)
	    a_ptr[a0_bas+a0_stp*s+a1_off]/=d;
	  
	} j.ta();

	//	System.out.println("iter "+k);

	Adlib.remap(b.section(x.triplet(k+1, n-1)), 
		    a.section(x.triplet(k+1, n-1),y.location(k)));

	LocBlocksIndex m = new LocBlocksIndex(y.triplet(k+1, n-1));

	//over

	for(m.beginLocBlk(); m.test(); m.nextLocBlk()) {

	  final int a1_bas=a.stride[1].offset(m);
	  final int a1_stp=a.stride[1].step(m);

	  final int m_count=m.count();
	  final int m_glb_bas=m.glb_bas();
	  final int m_glb_stp=m.glb_stp();

	  int m_idx=k+1+m_glb_bas;

	  for (int m_blk=0; m_blk<m_count; m_blk++) {
	    
	    //final int m_idx=k+1+m_glb_bas+m_glb_stp*m_blk;
	    final int b_idx=b0_bas+b0_stp*m_idx;

	    int a_bas=a0_bas+a0_stp*m_idx+a1_bas+a1_stp*m_blk;
	    int b_bas=b0_bas+b0_stp*m_idx;

 	    for (int ii=m_idx; ii<n; ii++) {
 	      a_ptr[a_bas]-=b_ptr[b_bas] * b_ptr[b_idx];
	      a_bas+=a0_stp;
	      b_bas+=b0_stp;
 	    }

// 	    for (int ii=m_idx; ii<n; ii++) {
// 	      a_ptr[a0_bas+a0_stp*ii+a1_bas+a1_stp*m_blk]-=
// 		b_ptr[b0_bas+b0_stp*ii] *
// 		b_ptr[b0_bas+b0_stp*m_idx];
// 	    }

	    m_idx+=m_glb_stp;

	  }

	}

	//revo
	
	//	Adlib.rprintln("A = ");
	//	Adlib.aprintf("%7.3f%"+n+"N", a);
      }

      j=y.location(n-1);

      if (j.at()) {
	final int a1_off=a.stride[1].offset(j);

	a_ptr[a0_bas+a0_stp*(n-1)+a1_off]=
	  (float)Math.sqrt(a_ptr[a0_bas+a0_stp*(n-1)+a1_off]);

      } j.ta();

      Clock.stop();

      //Adlib.rprintln("A = ");
      //Adlib.aprintf("%7.3f%"+n+"N", a);
    
    } p.no();

    spmd.finalize();

  }
}
